@@ -50,6 +50,14 @@ class DeferredStrategy(SearchStrategy):
         )
         self.__settings = settings.default or settings()
 
+    @property
+    def supports_find(self):
+        return self.wrapped_strategy.supports_find
+
+    @property
+    def is_empty(self):
+        return self.wrapped_strategy.is_empty
+
     @property
     def wrapped_strategy(self):
         if self.__wrapped_strategy is None:
@@ -61,11 +69,10 @@ class DeferredStrategy(SearchStrategy):
         return self.__wrapped_strategy
 
     def validate(self):
-        self.wrapped_strategy.validate()
-
-    @property
-    def template_upper_bound(self):
-        return self.wrapped_strategy.template_upper_bound
+        w = self.wrapped_strategy
+        assert isinstance(w, SearchStrategy), \
+            '%r returned non-strategy %r' % (self, w)
+        w.validate()
 
     def __repr__(self):
         if self.__representation is None:
@@ -93,23 +100,5 @@ class DeferredStrategy(SearchStrategy):
             )
         return self.__representation
 
-    def draw_parameter(self, random):
-        return self.wrapped_strategy.draw_parameter(random)
-
-    def draw_template(self, random, pv):
-        return self.wrapped_strategy.draw_template(random, pv)
-
-    def reify(self, value):
-        return self.wrapped_strategy.reify(value)
-
-    def simplifiers(self, random, template):
-        return self.wrapped_strategy.simplifiers(random, template)
-
-    def strictly_simpler(self, x, y):
-        return self.wrapped_strategy.strictly_simpler(x, y)
-
-    def to_basic(self, template):
-        return self.wrapped_strategy.to_basic(template)
-
-    def from_basic(self, data):
-        return self.wrapped_strategy.from_basic(data)
+    def do_draw(self, data):
+        return data.draw(self.wrapped_strategy)
