@@ -17,63 +17,125 @@
 from __future__ import division, print_function, absolute_import
 
 import math
-import struct
 from decimal import Decimal
 
 from hypothesis.errors import InvalidArgument
 from hypothesis.control import assume
-from hypothesis.settings import Settings
 from hypothesis.searchstrategy import SearchStrategy
-from hypothesis.internal.compat import hrange, ArgSpec, text_type, \
-    getargspec, integer_types, float_to_decimal, unicode_safe_repr
+from hypothesis.internal.compat import ArgSpec, text_type, getargspec, \
+    integer_types, float_to_decimal
+from hypothesis.internal.floats import is_negative, float_to_int, \
+    int_to_float, count_between_floats
+from hypothesis.utils.conventions import not_set
+from hypothesis.internal.reflection import proxies
 from hypothesis.searchstrategy.reprwrapper import ReprWrapperStrategy
 
 __all__ = [
+    'nothing',
     'just', 'one_of',
     'none',
-    'choices',
+    'choices', 'streaming',
     'booleans', 'integers', 'floats', 'complex_numbers', 'fractions',
     'decimals',
     'characters', 'text', 'binary',
     'tuples', 'lists', 'sets', 'frozensets',
     'dictionaries', 'fixed_dictionaries',
-    'sampled_from',
+    'sampled_from', 'permutations',
     'builds',
-    'randoms',
-    'streaming', 'basic', 'recursive', 'composite',
+    'randoms', 'random_module',
+    'recursive', 'composite',
+    'shared', 'runner',
+    'recursive', 'composite',
 ]
 
+_strategies = set()
+
+
+class FloatKey(object):
+
+    def __init__(self, f):
+        self.value = float_to_int(f)
+
+    def __eq__(self, other):
+        return isinstance(other, FloatKey) and (
+            other.value == self.value
+        )
+
+    def __ne__(self, other):
+        return not self.__eq__(other)
+
+    def __hash__(self):
+        return hash(self.value)
+
+
+def convert_value(v):
+    if isinstance(v, float):
+        return FloatKey(v)
+    return (type(v), v)
+
+
+def cacheable(fn):
+    cache = {}
+
+    @proxies(fn)
+    def cached_strategy(*args, **kwargs):
+        kwargs_cache_key = set()
+        try:
+            for k, v in kwargs.items():
+                kwargs_cache_key.add((k, convert_value(v)))
+        except TypeError:
+            return fn(*args, **kwargs)
+        cache_key = (
+            tuple(map(convert_value, args)), frozenset(kwargs_cache_key))
+        try:
+            return cache[cache_key]
+        except TypeError:
+            return fn(*args, **kwargs)
+        except KeyError:
+            result = fn(*args, **kwargs)
+            cache[cache_key] = result
+            return result
+    return cached_strategy
+
 
 def defines_strategy(strategy_definition):
-    from hypothesis.internal.reflection import proxies, arg_string, \
-        convert_positional_arguments
-    argspec = getargspec(strategy_definition)
-    defaults = {}
-    if argspec.defaults is not None:
-        for k in hrange(1, len(argspec.defaults) + 1):
-            defaults[argspec.args[-k]] = argspec.defaults[-k]
+    from hypothesis.searchstrategy.deferred import DeferredStrategy
+    _strategies.add(strategy_definition.__name__)
 
     @proxies(strategy_definition)
     def accept(*args, **kwargs):
-        result = strategy_definition(*args, **kwargs)
-
-        def calc_repr():
-            _args = args
-            _kwargs = kwargs
-            _args, _kwargs = convert_positional_arguments(
-                strategy_definition, _args, _kwargs)
-            kwargs_for_repr = dict(_kwargs)
-            for k, v in defaults.items():
-                if k in kwargs_for_repr and kwargs_for_repr[k] is defaults[k]:
-                    del kwargs_for_repr[k]
-            return u'%s(%s)' % (
-                strategy_definition.__name__,
-                arg_string(strategy_definition, _args, kwargs_for_repr)
-            )
-        return ReprWrapperStrategy(result, calc_repr)
+        return DeferredStrategy(strategy_definition, args, kwargs)
     return accept
 
 
+class Nothing(SearchStrategy):
+    is_empty = True
+
+    def do_draw(self, data):
+        data.mark_invalid()
+
+    def __repr__(self):
+        return 'nothing()'
+
+    def map(self, f):
+        return self
+
+    def filter(self, f):
+        return self
+
+    def flatmap(self, f):
+        return self
+
+NOTHING = Nothing()
+
+
+@cacheable
+def nothing():
+    """This strategy never successfully draws a value and will always reject on
+    an attempt to draw."""
+    return NOTHING
+
+
 def just(value):
     """Return a strategy which only generates value.
 
@@ -83,7 +145,7 @@ def just(value):
     from hypothesis.searchstrategy.misc import JustStrategy
 
     def calc_repr():
-        return u'just(%s)' % (unicode_safe_repr(value),)
+        return 'just(%s)' % (repr(value),)
 
     return ReprWrapperStrategy(JustStrategy(value), calc_repr)
 
@@ -94,20 +156,33 @@ def none():
     return just(None)
 
 
-def one_of(arg, *args):
+def one_of(*args):
     """Return a strategy which generates values from any of the argument
-    strategies."""
+    strategies.
+
+    This may be called with one iterable argument instead of multiple strategy
+    arguments. In which case one_of(x) and one_of(\*x) are equivalent.
+
+    """
+    if len(args) == 1 and not isinstance(args[0], SearchStrategy):
+        try:
+            args = tuple(args[0])
+        except TypeError:
+            pass
 
-    if not args:
-        check_strategy(arg)
-        return arg
-    from hypothesis.searchstrategy.strategies import OneOfStrategy
-    args = (arg,) + args
     for arg in args:
         check_strategy(arg)
+    args = [a for a in args if not a.is_empty]
+
+    if not args:
+        return nothing()
+    if len(args) == 1:
+        return args[0]
+    from hypothesis.searchstrategy.strategies import OneOfStrategy
     return OneOfStrategy(args)
 
 
+@cacheable
 @defines_strategy
 def integers(min_value=None, max_value=None):
     """Returns a strategy which generates integers (in Python 2 these may be
@@ -118,33 +193,39 @@ def integers(min_value=None, max_value=None):
 
     """
 
+    check_valid_integer(min_value)
+    check_valid_integer(max_value)
+    check_valid_interval(min_value, max_value, 'min_value', 'max_value')
+
     from hypothesis.searchstrategy.numbers import IntegersFromStrategy, \
-        BoundedIntStrategy, RandomGeometricIntStrategy, WideRangeIntStrategy
+        BoundedIntStrategy, WideRangeIntStrategy
 
     if min_value is None:
         if max_value is None:
             return (
-                RandomGeometricIntStrategy() |
                 WideRangeIntStrategy()
             )
         else:
-            check_type(integer_types, max_value)
             return IntegersFromStrategy(0).map(lambda x: max_value - x)
     else:
-        check_type(integer_types, min_value)
         if max_value is None:
             return IntegersFromStrategy(min_value)
         else:
+            assert min_value <= max_value
             if min_value == max_value:
                 return just(min_value)
-            elif min_value > max_value:
-                raise InvalidArgument(
-                    u'Cannot have max_value=%r < min_value=%r' % (
-                        max_value, min_value
-                    ))
-            return BoundedIntStrategy(min_value, max_value)
+            elif min_value >= 0:
+                return BoundedIntStrategy(min_value, max_value)
+            elif max_value <= 0:
+                return BoundedIntStrategy(-max_value, -min_value).map(
+                    lambda t: -t
+                )
+            else:
+                return integers(min_value=0, max_value=max_value) | \
+                    integers(min_value=min_value, max_value=0)
 
 
+@cacheable
 @defines_strategy
 def booleans():
     """Returns a strategy which generates instances of bool."""
@@ -152,47 +233,37 @@ def booleans():
     return BoolStrategy()
 
 
-def is_negative(x):
-    return math.copysign(1, x) < 0
-
-
-def count_between_floats(x, y):
-    assert x <= y
-    if is_negative(x):
-        if is_negative(y):
-            return float_to_int(x) - float_to_int(y) + 1
-        else:
-            return count_between_floats(x, -0.0) + count_between_floats(0.0, y)
-    else:
-        assert not is_negative(y)
-        return float_to_int(y) - float_to_int(x) + 1
-
-
-def float_to_int(value):
-    return (
-        struct.unpack(b'!Q', struct.pack(b'!d', value))[0]
-    )
-
-
-def int_to_float(value):
-    return (
-        struct.unpack(b'!d', struct.pack(b'!Q', value))[0]
-    )
-
-
+@cacheable
 @defines_strategy
-def floats(min_value=None, max_value=None):
-    """Returns a strategy which generates floats. If min_value is not None,
-    all values will be >= min_value. If max_value is not None, all values will
-    be <= max_value.
+def floats(
+    min_value=None, max_value=None, allow_nan=None, allow_infinity=None
+):
+    """Returns a strategy which generates floats.
+
+    - If min_value is not None, all values will be >= min_value.
+    - If max_value is not None, all values will be <= max_value.
+    - If min_value or max_value is not None, it is an error to enable
+      allow_nan.
+    - If both min_value and max_value are not None, it is an error to enable
+      allow_infinity.
 
     Where not explicitly ruled out by the bounds, all of infinity, -infinity
     and NaN are possible values generated by this strategy.
+
     """
 
-    for e in (min_value, max_value):
-        if e is not None and math.isnan(e):
-            raise InvalidArgument(u'nan is not a valid end point')
+    if allow_nan is None:
+        allow_nan = bool(min_value is None and max_value is None)
+    elif allow_nan:
+        if min_value is not None or max_value is not None:
+            raise InvalidArgument(
+                'Cannot have allow_nan=%r, with min_value or max_value' % (
+                    allow_nan
+                ))
+
+    check_valid_bound(min_value, 'min_value')
+    check_valid_bound(max_value, 'max_value')
+    check_valid_interval(min_value, max_value, 'min_value', 'max_value')
     if min_value is not None:
         min_value = float(min_value)
     if max_value is not None:
@@ -202,26 +273,24 @@ def floats(min_value=None, max_value=None):
     if max_value == float(u'inf'):
         max_value = None
 
-    from hypothesis.searchstrategy.numbers import WrapperFloatStrategy, \
-        GaussianFloatStrategy, BoundedFloatStrategy, ExponentialFloatStrategy,\
-        JustIntFloats, NastyFloats, FullRangeFloats, \
+    if allow_infinity is None:
+        allow_infinity = bool(min_value is None or max_value is None)
+    elif allow_infinity:
+        if min_value is not None and max_value is not None:
+            raise InvalidArgument(
+                'Cannot have allow_infinity=%r, with both min_value and '
+                'max_value' % (
+                    allow_infinity
+                ))
+
+    from hypothesis.searchstrategy.numbers import FloatStrategy, \
         FixedBoundedFloatStrategy
     if min_value is None and max_value is None:
-        return WrapperFloatStrategy(
-            GaussianFloatStrategy() |
-            BoundedFloatStrategy() |
-            ExponentialFloatStrategy() |
-            JustIntFloats() |
-            NastyFloats() |
-            FullRangeFloats()
+        return FloatStrategy(
+            allow_infinity=allow_infinity, allow_nan=allow_nan,
         )
     elif min_value is not None and max_value is not None:
-        if max_value < min_value:
-            raise InvalidArgument(
-                u'Cannot have max_value=%r < min_value=%r' % (
-                    max_value, min_value
-                ))
-        elif min_value == max_value:
+        if min_value == max_value:
             return just(min_value)
         elif math.isinf(max_value - min_value):
             assert min_value < 0 and max_value > 0
@@ -229,16 +298,9 @@ def floats(min_value=None, max_value=None):
                 min_value=min_value, max_value=0
             )
         elif count_between_floats(min_value, max_value) > 1000:
-            critical_values = [
-                min_value, max_value, min_value + (max_value - min_value) / 2]
-            if min_value <= 0 <= max_value:
-                if not is_negative(max_value):
-                    critical_values.append(0.0)
-                if is_negative(min_value):
-                    critical_values.append(-0.0)
             return FixedBoundedFloatStrategy(
                 lower_bound=min_value, upper_bound=max_value
-            ) | sampled_from(critical_values)
+            )
         elif is_negative(max_value):
             assert is_negative(min_value)
             ub_int = float_to_int(max_value)
@@ -259,30 +321,38 @@ def floats(min_value=None, max_value=None):
                 int_to_float
             )
     elif min_value is not None:
-        critical_values = [min_value, float(u'inf')]
-        if is_negative(min_value):
-            critical_values.append(-0.0)
-        if min_value <= 0:
-            critical_values.append(0.0)
-        return (
-            floats().map(
-                lambda x: assume(not math.isnan(x)) and min_value + abs(x)
+        if min_value < 0:
+            result = floats(
+                min_value=0.0
+            ) | floats(min_value=min_value, max_value=0.0)
+        else:
+            result = (
+                floats(allow_infinity=allow_infinity, allow_nan=False).map(
+                    lambda x: assume(not math.isnan(x)) and min_value + abs(x)
+                )
             )
-        ) | sampled_from(critical_values)
+        if min_value == 0 and not is_negative(min_value):
+            result = result.filter(lambda x: math.copysign(1.0, x) == 1)
+        return result
     else:
         assert max_value is not None
-        critical_values = [max_value, float(u'-inf')]
-        if max_value >= 0:
-            critical_values.append(-0.0)
-            if not is_negative(max_value):
-                critical_values.append(0.0)
-        return (
-            floats().map(
-                lambda x: assume(not math.isnan(x)) and max_value - abs(x)
+        if max_value > 0:
+            result = floats(
+                min_value=0.0,
+                max_value=max_value,
+            ) | floats(max_value=0.0)
+        else:
+            result = (
+                floats(allow_infinity=allow_infinity, allow_nan=False).map(
+                    lambda x: assume(not math.isnan(x)) and max_value - abs(x)
+                )
             )
-        ) | sampled_from(critical_values)
+        if max_value == 0 and is_negative(max_value):
+            result = result.filter(is_negative)
+        return result
 
 
+@cacheable
 @defines_strategy
 def complex_numbers():
     """Returns a strategy that generates complex numbers."""
@@ -292,6 +362,7 @@ def complex_numbers():
     )
 
 
+@cacheable
 @defines_strategy
 def tuples(*args):
     """Return a strategy which generates a tuple of the same length as args by
@@ -303,10 +374,15 @@ def tuples(*args):
     """
     for arg in args:
         check_strategy(arg)
+
+    for arg in args:
+        if arg.is_empty:
+            return nothing()
     from hypothesis.searchstrategy.collections import TupleStrategy
     return TupleStrategy(args, tuple)
 
 
+@defines_strategy
 def sampled_from(elements):
     """Returns a strategy which generates any value present in the iterable
     elements.
@@ -320,23 +396,14 @@ def sampled_from(elements):
         JustStrategy
     elements = tuple(iter(elements))
     if not elements:
-        raise InvalidArgument(
-            u'sampled_from requires at least one value'
-        )
+        return nothing()
     if len(elements) == 1:
-        result = JustStrategy(elements[0])
+        return JustStrategy(elements[0])
     else:
-        result = SampledFromStrategy(elements)
-
-    def calc_repr():
-        return u'sampled_from((%s%s))' % (
-            u', '.join(map(unicode_safe_repr, elements)),
-            ',' if len(elements) == 1 else '',
-        )
-
-    return ReprWrapperStrategy(result, calc_repr)
+        return SampledFromStrategy(elements)
 
 
+@cacheable
 @defines_strategy
 def lists(
     elements=None, min_size=None, average_size=None, max_size=None,
@@ -360,29 +427,36 @@ def lists(
     condition that for i != j, unique_by(result[i]) != unique_by(result[j]).
 
     """
+    check_valid_sizes(min_size, average_size, max_size)
+    if elements is None or (max_size is not None and max_size <= 0):
+        if max_size is None or max_size > 0:
+            raise InvalidArgument(
+                u'Cannot create non-empty lists without an element type'
+            )
+        else:
+            return builds(list)
+    check_strategy(elements)
+    if elements.is_empty:
+        if (min_size or 0) > 0:
+            raise InvalidArgument((
+                'Cannot create non-empty lists with elements drawn from '
+                'strategy %r because it has no values.') % (elements,))
+        else:
+            return builds(list)
     if unique:
         if unique_by is not None:
             raise InvalidArgument((
-                u'cannot specify both unique and unique_by (you probably only '
-                u'want to set unique_by)'
+                'cannot specify both unique and unique_by (you probably only '
+                'want to set unique_by)'
             ))
         else:
             unique_by = lambda x: x
 
     if unique_by is not None:
         from hypothesis.searchstrategy.collections import UniqueListStrategy
-        if max_size == 0:
-            return builds(list)
         check_strategy(elements)
-        if min_size is not None and elements.template_upper_bound < min_size:
-            raise InvalidArgument((
-                u'Cannot generate unique lists of size %d from %r, which '
-                u'contains no more than %d distinct values') % (
-                    min_size, elements, elements.template_upper_bound,
-            ))
         min_size = min_size or 0
         max_size = max_size or float(u'inf')
-        max_size = min(max_size, elements.template_upper_bound)
         if average_size is None:
             if max_size < float(u'inf'):
                 if max_size <= 5:
@@ -391,7 +465,7 @@ def lists(
                     average_size = (max_size + min_size) / 2
             else:
                 average_size = max(
-                    Settings.default.average_list_length,
+                    _AVERAGE_LIST_LENGTH,
                     min_size * 2
                 )
         check_valid_sizes(min_size, average_size, max_size)
@@ -405,39 +479,23 @@ def lists(
         return result
 
     check_valid_sizes(min_size, average_size, max_size)
-    from hypothesis.searchstrategy.collections import ListStrategy, \
-        SingleElementListStrategy
+    from hypothesis.searchstrategy.collections import ListStrategy
     if min_size is None:
         min_size = 0
     if average_size is None:
         if max_size is None:
-            average_size = Settings.default.average_list_length
+            average_size = _AVERAGE_LIST_LENGTH
         else:
             average_size = (min_size + max_size) * 0.5
 
-    if elements is None or (max_size is not None and max_size <= 0):
-        if max_size is None or max_size > 0:
-            raise InvalidArgument(
-                u'Cannot create non-empty lists without an element type'
-            )
-        else:
-            return ListStrategy(())
-    else:
-        check_strategy(elements)
-        if elements.template_upper_bound == 1:
-            from hypothesis.searchstrategy.numbers import IntegersFromStrategy
-            if max_size is None:
-                length_strat = IntegersFromStrategy(
-                    min_size, average_size=average_size - min_size)
-            else:
-                length_strat = integers(min_size, max_size)
-            return SingleElementListStrategy(elements, length_strat)
-        return ListStrategy(
-            (elements,), average_length=average_size,
-            min_size=min_size, max_size=max_size,
-        )
+    check_strategy(elements)
+    return ListStrategy(
+        (elements,), average_length=average_size,
+        min_size=min_size, max_size=max_size,
+    )
 
 
+@cacheable
 @defines_strategy
 def sets(elements=None, min_size=None, average_size=None, max_size=None):
     """This has the same behaviour as lists, but returns sets instead.
@@ -453,6 +511,7 @@ def sets(elements=None, min_size=None, average_size=None, max_size=None):
     ).map(set)
 
 
+@cacheable
 @defines_strategy
 def frozensets(elements=None, min_size=None, average_size=None, max_size=None):
     """This is identical to the sets function but instead returns
@@ -477,10 +536,14 @@ def fixed_dictionaries(mapping):
     from hypothesis.searchstrategy.collections import FixedKeysDictStrategy
     check_type(dict, mapping)
     for v in mapping.values():
-        check_type(SearchStrategy, v)
+        check_strategy(v)
+    for v in mapping.values():
+        if v.is_empty:
+            return nothing()
     return FixedKeysDictStrategy(mapping)
 
 
+@cacheable
 @defines_strategy
 def dictionaries(
     keys, values, dict_class=dict,
@@ -498,18 +561,6 @@ def dictionaries(
     check_strategy(keys)
     check_strategy(values)
 
-    if min_size is not None and min_size > keys.template_upper_bound:
-        raise InvalidArgument((
-            u'Cannot generate dictionaries of size %d with keys from %r, '
-            u'which contains no more than %d distinct values') % (
-                min_size, keys, keys.template_upper_bound,
-        ))
-
-    if max_size is None:
-        max_size = keys.template_upper_bound
-    else:
-        max_size = min(max_size, keys.template_upper_bound)
-
     return lists(
         tuples(keys, values),
         min_size=min_size, average_size=average_size, max_size=max_size,
@@ -517,6 +568,7 @@ def dictionaries(
     ).map(dict_class)
 
 
+@cacheable
 @defines_strategy
 def streaming(elements):
     """Generates an infinite stream of values where each value is drawn from
@@ -531,6 +583,7 @@ def streaming(elements):
     return StreamStrategy(elements)
 
 
+@cacheable
 @defines_strategy
 def characters(whitelist_categories=None, blacklist_categories=None,
                blacklist_characters=None, min_codepoint=None,
@@ -549,6 +602,16 @@ def characters(whitelist_categories=None, blacklist_categories=None,
     pass them with `blacklist_characters` argument.
 
     """
+    if (
+        min_codepoint is not None and max_codepoint is not None and
+        min_codepoint > max_codepoint
+    ):
+        raise InvalidArgument(
+            'Cannot have min_codepoint=%d > max_codepoint=%d ' % (
+                min_codepoint, max_codepoint
+            )
+        )
+
     from hypothesis.searchstrategy.strings import OneCharStringStrategy
     return OneCharStringStrategy(whitelist_categories=whitelist_categories,
                                  blacklist_categories=blacklist_categories,
@@ -557,6 +620,7 @@ def characters(whitelist_categories=None, blacklist_categories=None,
                                  max_codepoint=max_codepoint)
 
 
+@cacheable
 @defines_strategy
 def text(
     alphabet=None,
@@ -571,14 +635,13 @@ def text(
     min_size, max_size and average_size have the usual interpretations.
 
     """
-    from hypothesis.searchstrategy.strings import OneCharStringStrategy, \
-        StringStrategy
+    from hypothesis.searchstrategy.strings import StringStrategy
     if alphabet is None:
-        char_strategy = OneCharStringStrategy(blacklist_categories=['Cs'])
+        char_strategy = characters(blacklist_categories=('Cs',))
     elif not alphabet:
         if (min_size or 0) > 0:
             raise InvalidArgument(
-                u'Invalid min_size %r > 0 for empty alphabet' % (
+                'Invalid min_size %r > 0 for empty alphabet' % (
                     min_size,
                 )
             )
@@ -593,6 +656,7 @@ def text(
     ))
 
 
+@cacheable
 @defines_strategy
 def binary(
     min_size=None, average_size=None, max_size=None
@@ -603,7 +667,11 @@ def binary(
     min_size, average_size and max_size have the usual interpretations.
 
     """
-    from hypothesis.searchstrategy.strings import BinaryStringStrategy
+    from hypothesis.searchstrategy.strings import BinaryStringStrategy, \
+        FixedSizeBytes
+    check_valid_sizes(min_size, average_size, max_size)
+    if min_size == max_size is not None:
+        return FixedSizeBytes(min_size)
     return BinaryStringStrategy(
         lists(
             integers(min_value=0, max_value=255),
@@ -612,33 +680,7 @@ def binary(
     )
 
 
-@defines_strategy
-def basic(
-    basic=None,
-    generate_parameter=None, generate=None, simplify=None, copy=None
-):
-    """Provides a facility to write your own strategies with significantly less
-    work.
-
-    See documentation for more details.
-
-    """
-    from hypothesis.searchstrategy.basic import basic_strategy, BasicStrategy
-    from copy import deepcopy
-    if basic is not None:
-        if isinstance(basic, type):
-            basic = basic()
-        check_type(BasicStrategy, basic)
-        generate_parameter = generate_parameter or basic.generate_parameter
-        generate = generate or basic.generate
-        simplify = simplify or basic.simplify
-        copy = copy or basic.copy
-    return basic_strategy(
-        parameter=generate_parameter,
-        generate=generate, simplify=simplify, copy=copy or deepcopy
-    )
-
-
+@cacheable
 @defines_strategy
 def randoms():
     """Generates instances of Random (actually a Hypothesis specific
@@ -647,6 +689,45 @@ def randoms():
     return RandomStrategy(integers())
 
 
+class RandomSeeder(object):
+
+    def __init__(self, seed):
+        self.seed = seed
+
+    def __repr__(self):
+        return 'random.seed(%r)' % (self.seed,)
+
+
+@cacheable
+@defines_strategy
+def random_module():
+    """If your code depends on the global random module then you need to use
+    this.
+
+    It will explicitly seed the random module at the start of your test
+    so that tests are reproducible. The value it passes you is an opaque
+    object whose only useful feature is that its repr displays the
+    random seed. It is not itself a random number generator. If you want
+    a random number generator you should use the randoms() strategy
+    which will give you one.
+
+    """
+    from hypothesis.control import cleanup
+    import random
+
+    def seed_random(seed):
+        state = random.getstate()
+        random.seed(seed)
+        cleanup(lambda: random.setstate(state))
+        return RandomSeeder(seed)
+
+    return shared(
+        integers().map(seed_random),
+        'hypothesis.strategies.random_module()',
+    )
+
+
+@cacheable
 @defines_strategy
 def fractions():
     """Generates instances of fractions.Fraction."""
@@ -656,6 +737,7 @@ def fractions():
     )
 
 
+@cacheable
 @defines_strategy
 def decimals():
     """Generates instances of decimals.Decimal."""
@@ -667,6 +749,8 @@ def decimals():
     )
 
 
+@cacheable
+@defines_strategy
 def builds(target, *args, **kwargs):
     """Generates values by drawing from args and kwargs and passing them to
     target in the appropriate argument position.
@@ -676,26 +760,10 @@ def builds(target, *args, **kwargs):
     call target(i, flag=b).
 
     """
-    from hypothesis.internal.reflection import nicerepr
-
-    def splat(value):
-        return target(*value[0], **value[1])
-    target_name = getattr(target, u'__name__', type(target).__name__)
-    splat.__name__ = str(
-        u'splat(%s)' % (target_name,)
+    return tuples(tuples(*args), fixed_dictionaries(kwargs)).map(
+        lambda value: target(*value[0], **value[1])
     )
 
-    def calc_repr():
-        return u'builds(%s)' % (
-            u', '.join(
-                [nicerepr(target)] +
-                list(map(nicerepr, args)) +
-                sorted([u'%s=%r' % (k, v) for k, v in kwargs.items()])))
-
-    return ReprWrapperStrategy(
-        tuples(tuples(*args), fixed_dictionaries(kwargs)).map(splat),
-        calc_repr)
-
 
 @defines_strategy
 def recursive(base, extend, max_leaves=100):
@@ -721,7 +789,7 @@ def recursive(base, extend, max_leaves=100):
     extended = extend(base)
     if not isinstance(extended, SearchStrategy):
         raise InvalidArgument(
-            u'Expected extend(%r) to be a SearchStrategy but got %r' % (
+            'Expected extend(%r) to be a SearchStrategy but got %r' % (
                 base, extended
             ))
     from hypothesis.searchstrategy.recursive import RecursiveStrategy
@@ -746,6 +814,7 @@ def permutations(values):
     return lists(tuples(index, index), max_size=n ** 2).map(build_permutation)
 
 
+@cacheable
 def composite(f):
     """Defines a strategy that is built out of potentially arbitrarily many
     other strategies.
@@ -755,7 +824,6 @@ def composite(f):
 
     """
 
-    from hypothesis.searchstrategy.morphers import MorpherStrategy
     from hypothesis.internal.reflection import copy_argspec
     argspec = getargspec(f)
 
@@ -776,26 +844,20 @@ def composite(f):
         keywords=argspec.keywords, defaults=argspec.defaults
     )
 
-    base_strategy = streaming(MorpherStrategy())
-
     @defines_strategy
     @copy_argspec(f.__name__, new_argspec)
     def accept(*args, **kwargs):
-        def call_with_draw(morphers):
-            index = [0]
+        class CompositeStrategy(SearchStrategy):
 
-            def draw(strategy):
-                i = index[0]
-                index[0] += 1
-                return morphers[i].become(strategy)
-            return f(*((draw,) + args), **kwargs)
-        return base_strategy.map(call_with_draw)
+            def do_draw(self, data):
+                return f(data.draw, *args, **kwargs)
+        return CompositeStrategy()
     return accept
 
 
 def shared(base, key=None):
     """Returns a strategy that draws a single shared value per run, drawn from
-    base. Any two shared instances with the samme key will share the same
+    base. Any two shared instances with the same key will share the same
     value, otherwise the identity of this strategy will be used. That is:
 
     >>> x = shared(s)
@@ -811,6 +873,7 @@ def shared(base, key=None):
     return SharedStrategy(base, key)
 
 
+@cacheable
 def choices():
     """Strategy that generates a function that behaves like random.choice.
 
@@ -818,40 +881,41 @@ def choices():
 
     """
     from hypothesis.control import note, current_build_context
+    from hypothesis.internal.conjecture.utils import choice
 
-    def build_chooser(stream):
-        index = [-1]
-        choice_count = [0]
-        context = current_build_context()
-        context.mark_captured()
+    class Chooser(object):
 
-        def choice(values):
+        def __init__(self, build_context, data):
+            self.build_context = build_context
+            self.data = data
+            self.choice_count = 0
+
+        def __call__(self, values):
             if not values:
                 raise IndexError('Cannot choose from empty sequence')
-            k = len(values) - 1
-            if k == 0:
-                chosen = 0
-            else:
-                mask = _right_saturate(k)
-                while True:
-                    index[0] += 1
-                    probe = stream[index[0]] & mask
-                    if probe <= k:
-                        chosen = probe
-                        break
-            choice_count[0] += 1
-            result = values[chosen]
-            with context.local():
-                note('Choice #%d: %r' % (choice_count[0], result))
+            result = choice(self.data, values)
+            with self.build_context.local():
+                self.choice_count += 1
+                note('Choice #%d: %r' % (self.choice_count, result))
             return result
-        return choice
+
+        def __repr__(self):
+            return 'choice'
+
+    class ChoiceStrategy(SearchStrategy):
+        supports_find = False
+
+        def do_draw(self, data):
+            return Chooser(current_build_context(), data)
+
     return ReprWrapperStrategy(
         shared(
-            builds(build_chooser, streaming(integers(min_value=0))),
+            ChoiceStrategy(),
             key='hypothesis.strategies.chooser.choice_function'
-        ), 'chooser()')
+        ), 'choices()')
 
 
+@cacheable
 def uuids():
     """Returns a strategy that generates UUIDs.
 
@@ -866,35 +930,137 @@ def uuids():
         ), 'uuids()')
 
 
+@defines_strategy
+def runner(default=not_set):
+    """A strategy for getting "the current test runner", whatever that may be.
+    The exact meaning depends on the entry point, but it will usually be the
+    associated 'self' value for it.
+
+    If there is no current test runner and a default is provided, return
+    that default. If no default is provided, raises InvalidArgument.
+
+    """
+    class RunnerStrategy(SearchStrategy):
+
+        def do_draw(self, data):
+            runner = getattr(data, 'hypothesis_runner', not_set)
+            if runner is not_set:
+                if default is not_set:
+                    raise InvalidArgument(
+                        'Cannot use runner() strategy with no '
+                        'associated runner or explicit default.'
+                    )
+                else:
+                    return default
+            else:
+                return runner
+    return RunnerStrategy()
+
+
+@cacheable
+def data():
+    """This isn't really a normal strategy, but instead gives you an object
+    which can be used to draw data interactively from other strategies.
+
+    It can only be used within @given, not find. This is because the lifetime
+    of the object cannot outlast the test body.
+
+    See the rest of the documentation for more complete information.
+
+    """
+    from hypothesis.control import note
+
+    class DataObject(object):
+
+        def __init__(self, data):
+            self.count = 0
+            self.data = data
+
+        def __repr__(self):
+            return 'data(...)'
+
+        def draw(self, strategy):
+            result = self.data.draw(strategy)
+            self.count += 1
+            note('Draw %d: %r' % (self.count, result))
+            return result
+
+    class DataStrategy(SearchStrategy):
+        supports_find = False
+
+        def do_draw(self, data):
+            if not hasattr(data, 'hypothesis_shared_data_strategy'):
+                data.hypothesis_shared_data_strategy = DataObject(data)
+            return data.hypothesis_shared_data_strategy
+
+        def __repr__(self):
+            return 'data()'
+
+        def map(self, f):
+            self.__not_a_first_class_strategy('map')
+
+        def filter(self, f):
+            self.__not_a_first_class_strategy('filter')
+
+        def flatmap(self, f):
+            self.__not_a_first_class_strategy('flatmap')
+
+        def example(self):
+            self.__not_a_first_class_strategy('example')
+
+        def __not_a_first_class_strategy(self, name):
+            raise InvalidArgument((
+                'Cannot call %s on a DataStrategy. You should probably be '
+                "using @composite for whatever it is you're trying to do."
+            ) % (name,))
+    return DataStrategy()
+
 # Private API below here
 
 
-def _right_saturate(x):
-    x |= (x >> 1)
-    x |= (x >> 2)
-    x |= (x >> 4)
-    x |= (x >> 8)
-    x |= (x >> 16)
-    x |= (x >> 32)
-    return x
-
-
 def check_type(typ, arg):
     if not isinstance(arg, typ):
         if isinstance(typ, type):
             typ_string = typ.__name__
         else:
-            typ_string = u'one of %s' % (
-                u', '.join(t.__name__ for t in typ))
+            typ_string = 'one of %s' % (
+                ', '.join(t.__name__ for t in typ))
         raise InvalidArgument(
-            u'Expected %s but got %r' % (typ_string, arg,))
+            'Expected %s but got %r' % (typ_string, arg,))
 
 
 def check_strategy(arg):
     check_type(SearchStrategy, arg)
 
 
+def check_valid_integer(value):
+    """Checks that value is either unspecified, or a valid integer.
+
+    Otherwise raises InvalidArgument.
+
+    """
+    if value is None:
+        return
+    check_type(integer_types, value)
+
+
+def check_valid_bound(value, name):
+    """Checks that value is either unspecified, or a valid interval bound.
+
+    Otherwise raises InvalidArgument.
+
+    """
+    if value is None:
+        return
+    if math.isnan(value):
+        raise InvalidArgument(u'Invalid end point %s %r' % (value, name))
+
+
 def check_valid_size(value, name):
+    """Checks that value is either unspecified, or a valid non-negative size
+    expressed as an integer/float. Otherwise raises InvalidArgument.
+
+    """
     if value is None:
         return
     check_type(integer_types + (float,), value)
@@ -904,37 +1070,40 @@ def check_valid_size(value, name):
         raise InvalidArgument(u'Invalid size %s %r' % (value, name))
 
 
-def check_valid_sizes(min_size, average_size, max_size):
-    check_valid_size(min_size, u'min_size')
-    check_valid_size(max_size, u'max_size')
-    check_valid_size(average_size, u'average_size')
-    if max_size is not None:
-        if min_size is not None:
-            if max_size < min_size:
-                raise InvalidArgument(
-                    u'Cannot have max_size=%r < min_size=%r' % (
-                        max_size, min_size
-                    ))
+def check_valid_interval(lower_bound, upper_bound, lower_name, upper_name):
+    """Checks that lower_bound and upper_bound are either unspecified, or they
+    define a valid interval on the number line.
 
-        if average_size is not None:
-            if max_size < average_size:
-                raise InvalidArgument(
-                    u'Cannot have max_size=%r < average_size=%r' % (
-                        max_size, average_size
-                    ))
+    Otherwise raises InvalidArgument.
+
+    """
+    if lower_bound is None or upper_bound is None:
+        return
+    if upper_bound < lower_bound:
+        raise InvalidArgument(
+            'Cannot have %s=%r < %s=%r' % (
+                upper_name, upper_bound, lower_name, lower_bound
+            ))
+
+
+def check_valid_sizes(min_size, average_size, max_size):
+    check_valid_size(min_size, 'min_size')
+    check_valid_size(max_size, 'max_size')
+    check_valid_size(average_size, 'average_size')
+    check_valid_interval(min_size, max_size, 'min_size', 'max_size')
+    check_valid_interval(average_size, max_size, 'average_size', 'max_size')
+    check_valid_interval(min_size, average_size, 'min_size', 'average_size')
 
     if average_size is not None:
-        if min_size is not None:
-            if average_size < min_size:
-                raise InvalidArgument(
-                    u'Cannot have average_size=%r < min_size=%r' % (
-                        average_size, min_size
-                    ))
         if (
             (max_size is None or max_size > 0) and
             average_size is not None and average_size <= 0.0
         ):
             raise InvalidArgument(
-                u'Cannot have average_size=%r < min_size=%r' % (
+                'Cannot have average_size=%r < min_size=%r' % (
                     average_size, min_size
                 ))
+
+
+_AVERAGE_LIST_LENGTH = 5.0
+assert _strategies.issubset(set(__all__)), _strategies - set(__all__)
